from __future__ import annotations

from datetime import datetime
from pathlib import Path
from typing import (
    Any,
    ClassVar,
    Generic,
    Iterator,
    Type,
    TypeVar,
)
from uuid import uuid4

from pydantic import (
    BaseModel,
    Field,
)

from agentlens.exceptions import DatasetIndexError
from agentlens.lens import Lens

MISSING_LABEL = object()
SUBSET_FILTER_FN_INDICATOR = "_is_subset_filter_function"


def Label(
    default: Any = MISSING_LABEL,
    **kwargs,
) -> Any:
    return Field(default=default, **kwargs)


E = TypeVar("E", bound=BaseModel)


class DatasetMetadata(BaseModel):
    version: int
    created_at: datetime
    last_modified: datetime


class DatasetFile(BaseModel, Generic[E]):
    metadata: DatasetMetadata
    examples: list[E]


class Dataset(Generic[E]):
    example: ClassVar[Type[E]]  # type: ignore[misc]

    def __init__(
        self,
        name: str,
        lens: Lens,
        subset: str | None = None,
    ):
        self.name = name
        self.subset = subset
        self.dir = lens._dataset_dir
        self._lens = lens
        self._file = self._read_file()
        rows = self._file.examples
        if subset:
            subset_rows, _ = self.split_examples(rows)
            self.examples = subset_rows
        else:
            self.examples = rows

    def filter(self, example: E) -> bool:
        """May be overridden by subclasses"""
        return True

    def __init_subclass__(cls, **kwargs):
        """Initialize subclass with the example type"""

        super().__init_subclass__(**kwargs)
        for base in cls.__orig_bases__:
            if getattr(base, "__origin__", None) is Dataset:
                args = base.__args__
                if args:
                    cls.example = args[0]
                    break
        if cls.example is None:
            raise ValueError("Dataset must specify a row type")

    @property
    def file_path(self) -> Path:
        # stub = f".{self.subset}" if self.subset else ""
        stub = ""
        path = self.dir / f"{self.name}{stub}.json"
        path.parent.mkdir(parents=True, exist_ok=True)
        return path

    def _read_file(self) -> DatasetFile[E]:
        if not self.file_path.exists():
            return DatasetFile(
                metadata=DatasetMetadata(
                    version=1,
                    created_at=datetime.now(),
                    last_modified=datetime.now(),
                ),
                examples=[],
            )
        json_str = self.file_path.read_text()
        return DatasetFile[self.example].model_validate_json(json_str)  # type: ignore[name-defined]

    def split_examples(self, examples: list[E]) -> tuple[list[E], list[E]]:
        return (
            [x for x in examples if self.filter(x)],
            [x for x in examples if not self.filter(x)],
        )

    def extend(self, examples: list[E]) -> None:
        """Appends examples to the dataset in-place"""
        self.examples.extend(examples)

    def clear(self) -> None:
        """Clears the dataset in-place"""
        self.examples.clear()

    def save(self) -> None:
        """
        Saves the current version of the subset to disk, leaving other rows unchanged.
        """
        # concatenate current subset's rows with the rest of the dataset's rows
        new_rows = self.examples
        if self.subset:
            _, other_examples = self.split_examples(self._file.examples)
            new_rows.extend(other_examples)

        # overwrite current dataset with new rows
        self._file.examples = new_rows
        json_str = self._file.model_dump_json(indent=2, exclude_defaults=True)
        self.file_path.write_text(json_str)

    def __iter__(self) -> Iterator[E]:
        return iter(self.examples)

    def __getitem__(self, key: int) -> E:
        if 0 <= key < len(self.examples):
            return self.examples[key]
        raise DatasetIndexError(f"Row index {key} out of range")

    def __len__(self):
        return len(self.examples)


class Example(BaseModel):
    """
    A row in a Dataset. Will not raise validation errors if fields designated
    as Labels are missing
    """

    id: str = Field(default_factory=lambda: str(uuid4()))

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)
        if attr is MISSING_LABEL:
            raise AttributeError(f"Missing label: {name}")
        return attr

    # TODO: implement __setattr__ with validation at set-time instead of save-time



from __future__ import annotations

import inspect
from contextlib import contextmanager
from contextvars import ContextVar
from datetime import datetime
from pathlib import Path
from typing import Any, Callable, Generic, Iterator, Literal, Sequence, TypeVar
from uuid import UUID, uuid4

import petname
from pydantic import BaseModel, Field

from agentlens.hooks import Hook
from agentlens.utils import now

T = TypeVar("T")

# At module level
_do_mock: ContextVar[bool] = ContextVar("do_mock", default=False)
_do_cache: ContextVar[bool] = ContextVar("do_cache", default=False)


class Log(BaseModel):
    message: str
    timestamp: datetime = Field(default_factory=now)


class File(BaseModel):
    name: str
    content: str
    timestamp: datetime = Field(default_factory=now)


class Observation(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    name: str
    error: str | None = None
    start_time: datetime = Field(default_factory=now)
    end_time: datetime | None = None
    metadata: dict[str, Any] = Field(default_factory=dict)
    children: list[Observation] = Field(default_factory=list)
    logs: list[Log] = Field(default_factory=list)
    files: list[File] = Field(default_factory=list)

    def add_child(self, child: Observation) -> None:
        self.children.append(child)

    def end(self) -> None:
        self.end_time = now()

    def get_status(self) -> Literal["running", "completed", "failed"]:
        if self.error is not None:
            return "failed"
        if self.end_time is None:
            return "running"
        return "completed"

    def get_status_icon(self) -> Any:
        return {
            "running": "🔄",
            "completed": "✅",
            "failed": "❌",
        }[self.get_status()]

    def add_log(self, message: str) -> None:
        self.logs.append(Log(message=message))

    def add_file(self, name: str, content: str) -> None:
        self.files.append(File(name=name, content=content))


class Run(Generic[T]):
    def __init__(self, runs_dir: Path, name: str):
        self.key = self._create_run_key()
        self.dir = runs_dir / self.key
        self.dir.mkdir(parents=True, exist_ok=True)
        self.hooks: dict[str, list[Hook]] = {}
        self.observation = Observation(name=name)
        self.observation_stack: list[Observation] = [self.observation]

        # Context stores
        self.context_values: dict[str, Any] = {}

        # Initialize contextvars with default values
        self._mock_token = _do_mock.set(False)
        self._cache_token = _do_cache.set(False)

    @property
    def do_mock(self) -> bool:
        return _do_mock.get()

    @do_mock.setter
    def do_mock(self, value: bool) -> None:
        _do_mock.set(value)

    @property
    def do_cache(self) -> bool:
        return _do_cache.get()

    @do_cache.setter
    def do_cache(self, value: bool) -> None:
        _do_cache.set(value)

    def _create_run_key(self) -> str:
        timestamp = now().strftime("%Y%m%d_%H%M%S")
        id = petname.generate(words=3, separator="_")
        return f"{timestamp}_{id}"

    @contextmanager
    def create_observation(
        self,
        name: str,
        is_method: bool = False,
        func_args: tuple = (),
        func_kwargs: dict = {},
    ) -> Iterator[Observation]:
        stack = self.observation_stack.copy()
        if not stack:
            raise ValueError("Observation stack unexpectedly empty")
        parent = stack[-1]
        observation = Observation(name=name)
        parent.add_child(observation)
        self.observation_stack = stack + [observation]
        try:
            yield observation
        finally:
            observation.end()
            stack = self.observation_stack.copy()
            stack.pop()
            self.observation_stack = stack

    def initialize_hooks(
        self,
        example: T,
        hook_factories: Sequence[Callable[[T], Hook]] | None,
    ) -> dict[str, list[Hook]]:
        hooks: dict[str, list[Hook]] = {}
        for hook_factory in hook_factories or []:
            hook = hook_factory(example)
            target_name = hook.target.__name__
            if target_name not in hooks:
                hooks[target_name] = []
            hooks[target_name].append(hook)
        return hooks

    @staticmethod
    def _is_method(func: Callable) -> bool:
        params = inspect.signature(func).parameters
        return "self" in params or "cls" in params


class Generation(Observation):
    model: str
    prompt_tokens: int
    output_tokens: int
    cost: float



from contextlib import contextmanager
from contextvars import ContextVar
from functools import wraps
from logging import getLogger
from pathlib import Path
from typing import (
    Any,
    Awaitable,
    Callable,
    Coroutine,
    Generator,
    Iterable,
    Iterator,
    ParamSpec,
    Type,
    TypeVar,
    cast,
)

import tqdm
from tqdm.asyncio import tqdm_asyncio

from agentlens.hooks import GeneratorHook, Hook, Mock, MockMiss, convert_to_kwargs
from agentlens.trace import Run

_run_context: ContextVar[Run | None] = ContextVar("run_context", default=None)
_do_mock: ContextVar[bool] = ContextVar("do_mock", default=False)
_do_cache: ContextVar[bool] = ContextVar("do_cache", default=False)

F = TypeVar("F", bound=Callable[..., Any])
P = ParamSpec("P")
R = TypeVar("R", covariant=True)
T = TypeVar("T")
C = TypeVar("C")


class Lens:
    _log = getLogger("agentlens")
    _dataset_dir: Path
    _runs_dir: Path
    _context_types: dict[str, Type]
    _mock_registry: dict[Callable, Callable]

    def __init__(
        self,
        *,
        dataset_dir: Path | str,
        runs_dir: Path | str,
    ):
        self._dataset_dir = Path(dataset_dir)
        self._runs_dir = Path(runs_dir)
        self._context_types = {}
        self._mock_registry = {}

    def task(
        self,
        *,
        name: str | None = None,
        mock: Callable | None = None,
    ) -> (
        Callable[P, Coroutine[Any, Any, R]]
        | Callable[[Callable[P, Coroutine[Any, Any, R]]], Callable[P, Coroutine[Any, Any, R]]]
    ):
        def decorator(
            fn: Callable[P, Coroutine[Any, Any, R]],
        ) -> Callable[P, Coroutine[Any, Any, R]]:
            if name:
                fn.__name__ = name

            if mock is not None:
                setattr(fn, "_mock_fn", Mock(mock, fn))

            @wraps(fn)
            async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
                with self._provide_run(fn.__name__) as run:
                    with run.create_observation(fn.__name__):
                        indent = "    " * (len(run.observation_stack) - 1)
                        self._log.info(f"{indent}[{fn.__name__}]")

                        hooks = run.hooks.get(fn.__name__, [])

                        # run pre-hooks
                        generators: list[GeneratorHook] = []
                        injected_kwargs: dict[str, Any] = {}
                        for hook in hooks:
                            # should produce None or a generator
                            gen = hook(args, kwargs)
                            if isinstance(gen, Generator):
                                generators.append(gen)  # type: ignore[arg-type]
                                new_injected_kwargs = next(gen)
                                injected_kwargs.update(new_injected_kwargs)

                        # rewrite task args/kwargs
                        all_kwargs = convert_to_kwargs(fn, args, kwargs)
                        all_kwargs.update(injected_kwargs)

                        # execute task
                        mock_fn = getattr(fn, "_mock_fn", None)
                        try:
                            if run.do_mock and mock_fn is not None:
                                result = await mock_fn(**all_kwargs)
                            else:
                                raise MockMiss
                        except MockMiss:
                            result = await fn(**all_kwargs)

                        # send result to generator hooks
                        for gen in generators:
                            try:
                                gen.send(result)
                            except StopIteration:
                                pass

                        return result

            return wrapper

        return decorator

    def hook(self, target_fn: Callable[..., Awaitable[Any]]) -> Callable[[Callable], Hook]:
        def decorator(hook_fn: Callable) -> Hook:
            if not hasattr(hook_fn, "__name__"):
                raise ValueError("Hook function must have a __name__ attribute")
            return Hook(hook_fn, target_fn)

        return decorator

    @property
    def root(self) -> Path:
        run = self._get_current_run()
        return run.dir

    def _get_current_run(self) -> Run:
        run = _run_context.get()
        if run is None:
            raise ValueError("Attempted to access current Run outside of a run context")
        return run

    def __truediv__(self, other: str | Path) -> Path:
        return self.root / str(other)

    def iter(self, iterable: Iterable[T], desc: str | None = None) -> Iterator[T]:
        return tqdm.tqdm(iterable, desc=desc)

    async def gather(self, *coros: Awaitable[T], desc: str | None = None) -> list[T]:
        return await tqdm_asyncio.gather(*coros, desc=desc)

    def context(self, cls: Type[C]) -> Type[C]:
        key = cls.__name__
        if key in self._context_types:
            existing_cls = self._context_types[key]
            if existing_cls != cls:
                raise ValueError(
                    f"Context name '{key}' is already registered to {existing_cls.__name__}. "
                    f"Cannot register it again for {cls.__name__}"
                )

        # Register the context type
        self._context_types[key] = cls
        return cls

    @contextmanager
    def provide(
        self,
        *contexts: Any,
        hooks: list[Hook] = [],
    ) -> Generator[None, None, None]:
        with self._provide_run() as run:
            with self._provide_contexts(run, *contexts):
                with self._provide_hooks(run, hooks):
                    yield

    @contextmanager
    def _provide_run(self, task_name: str | None = None) -> Generator[Run, None, None]:
        """Manages the run context, creating a new one if needed."""
        run = _run_context.get()
        created_run = False

        if run is None:
            if task_name is None:
                task_name = "default"
            run = Run(self._runs_dir, task_name)
            _run_context.set(run)
            created_run = True

        try:
            yield run
        finally:
            if created_run:
                _run_context.set(None)

    @contextmanager
    def _provide_contexts(self, run: Run, *contexts: Any) -> Generator[None, None, None]:
        context_keys = [type(context).__name__ for context in contexts]

        # ensure that all contexts have been registered
        for key in context_keys:
            if key not in self._context_types:
                raise ValueError(
                    f"Type {key} has not been registered as a context. "
                    "Use the `@ls.context` decorator to register it."
                )

        # save previous values
        prev_values = {key: run.context_values.get(key) for key in context_keys}

        # set new values
        for context, key in zip(contexts, context_keys):
            run.context_values[key] = context

        try:
            yield
        finally:
            # restore previous values
            for key, prev_value in prev_values.items():
                if prev_value is not None:
                    run.context_values[key] = prev_value
                else:
                    del run.context_values[key]

    @contextmanager
    def _provide_hooks(self, run: Run, hooks: list[Hook]) -> Generator[None, None, None]:
        """Provide hooks for the current run"""
        # Group hooks by their target function name
        hook_map: dict[str, list[Hook]] = {}
        for hook in hooks:
            target_name = hook.target.__name__
            if target_name not in hook_map:
                hook_map[target_name] = []
            hook_map[target_name].append(hook)

        # Store previous hooks
        prev_hooks = run.hooks.copy()

        # Update run's hooks
        for target_name, target_hooks in hook_map.items():
            if target_name not in run.hooks:
                run.hooks[target_name] = []
            run.hooks[target_name].extend(target_hooks)

        try:
            yield
        finally:
            # Restore previous hooks
            run.hooks = prev_hooks

    def __getitem__(self, context_type: Type[T]) -> T:
        """Get a context value by its type"""
        key = context_type.__name__
        run = self._get_current_run()
        try:
            return cast(T, run.context_values[key])
        except KeyError:
            raise ValueError(
                f"No context value provided for type {context_type.__name__}. "
                "Use 'with ls.provide(value):' to provide one."
            )

    def eval(self):  # type: ignore[no-untyped-def]
        """Just an alias for task for the extension to consume"""
        return self.task

    @contextmanager
    def mock(self):
        """Request that tasks in this context use their mock implementations"""
        with self._provide_run() as run:
            prev_mock = run.do_mock
            run.do_mock = True
            try:
                yield
            finally:
                run.do_mock = prev_mock

    @contextmanager
    def no_mock(self):
        """Request that tasks in this context use their real implementations"""
        with self._provide_run() as run:
            prev_mock = run.do_mock
            run.do_mock = False
            try:
                yield
            finally:
                run.do_mock = prev_mock

    @contextmanager
    def cache(self):
        """Request that tasks in this context use cached results if available"""
        token = _do_cache.set(True)
        try:
            yield
        finally:
            _do_cache.reset(token)

    @contextmanager
    def no_cache(self):
        """Request that tasks in this context ignore cached results"""
        token = _do_cache.set(False)
        try:
            yield
        finally:
            _do_cache.reset(token)



from __future__ import annotations

import shutil
import subprocess
from logging import getLogger
from typing import Any, Callable, ClassVar

from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.widgets import Tree

from agentlens.trace import Observation, Run

logger = getLogger("agentlens.console")


class RunTree(Tree):
    COMPONENT_CLASSES: ClassVar[set[str]] = {"run-tree"}

    def __init__(self, runs: list[Run], editor: str = "cursor"):
        super().__init__("Runs", id="run-tree")
        self.runs = runs
        self.editor = editor

    def on_mount(self) -> None:
        self.show_root = False
        self.guide_depth = 3
        self.root.expand()

    def on_tree_node_selected(self, event: Tree.NodeSelected) -> None:
        node = event.node
        if node.data and node.data.get("type") == "file":
            path = node.data["path"]
            editor_path = shutil.which(self.editor)
            if not editor_path:
                raise Exception(f"Could not find '{self.editor}' in PATH")
            try:
                subprocess.Popen([editor_path, path])
            except Exception as e:
                raise Exception(f"Failed to open file in {self.editor}: {e}")

    def render_trace(self, trace: Observation, parent_node: Any, run: Run) -> None:
        status = trace.get_status()
        if status == "failed":
            color = "red3"
        elif status == "running":
            color = "dodger_blue1"
        else:
            color = "green3"

        node = parent_node.add(f"[b {color}]{trace.get_status_icon()} {trace.name}[/]", expand=True)

        for log in trace.logs:
            node.add_leaf(f"{log.message}")

        for file in trace.files:
            file_path = run.dir / file.name
            abs_path = str(file_path.resolve())
            leaf_node = node.add_leaf(f"📄 {file.name}")
            leaf_node.data = {"type": "file", "path": abs_path}

        for child in trace.children:
            self.render_trace(child, node, run)

    def refresh_tree(self) -> None:
        node_states = self._get_node_states(self.root)
        self.root.remove_children()

        for i, run in enumerate(self.runs):
            run_node = self.root.add(f"Run {i}", expand=True)
            self.render_trace(run.observation, run_node, run)

        self._apply_node_states(self.root, node_states)
        self.refresh()

    def _get_node_states(self, node, path=()):
        """Recursively get the expanded state of nodes."""
        states = {}
        for child in node.children:
            key = path + (str(child.label),)
            states[key] = child.is_expanded
            states.update(self._get_node_states(child, key))
        return states

    def _apply_node_states(self, node, states, path=()):
        """Recursively apply the expanded state to nodes."""
        for child in node.children:
            key = path + (str(child.label),)
            if key in states:
                if states[key]:
                    child.expand()
                else:
                    child.collapse()
            self._apply_node_states(child, states, key)


class RunConsole(App):
    CSS = """
    ScrollableContainer {
        width: 100%;
        height: 100%;
        dock: left;
        border: solid green;
    }

    .run-tree {
        width: 100%;
        height: 100%;
        background: #2c3e50;
        color: #ecf0f1;
        padding: 1;
    }
    """

    def __init__(
        self, runs: list[Run], execute_callback: Callable, editor: str = "cursor", *args, **kwargs
    ):
        super().__init__(*args, **kwargs)
        self._runs = runs
        self._execute_callback = execute_callback
        self._editor = editor

    def compose(self) -> ComposeResult:
        self.trace_tree = RunTree(self._runs, editor=self._editor)
        with ScrollableContainer():
            yield self.trace_tree

    async def on_mount(self) -> None:
        self.refresh_worker = self.set_interval(1 / 2, self.trace_tree.refresh_tree)
        self.execution_task = self.call_later(self.safe_execute)

    async def safe_execute(self) -> None:
        try:
            await self._execute_callback()
        except Exception as e:
            logger.error(f"Execution failed: {e}")



from __future__ import annotations

import json
import pickle
from abc import ABC, abstractmethod

from pydantic import BaseModel
from pydantic_core.core_schema import with_info_plain_validator_function


class Serializable(ABC):
    @abstractmethod
    def model_dump(self) -> dict: ...

    @classmethod
    @abstractmethod
    def model_validate(cls, value) -> Serializable: ...

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):
        return with_info_plain_validator_function(lambda value, info: cls.model_validate(value))

    def __eq__(self, other) -> bool:
        if not isinstance(other, Serializable):
            return NotImplemented
        return self.model_dump() == other.model_dump()

    def __hash__(self) -> int:
        return hash(json.dumps(self.model_dump(), sort_keys=True))

    def __getstate__(self):
        return self.model_dump()

    def __setstate__(self, state):
        obj = self.model_validate(state)
        self.__dict__.update(obj.__dict__)

    def test_serialized_to_json(self, dumped: dict):
        try:
            json.dumps(dumped)
        except Exception as e:
            raise ValueError(f"model_dump() returned a value that is not serializable to JSON: {e}")

    def test_serialization_roundtrip(self, dumped: dict):
        try:
            reconstructed = self.model_validate(dumped)
            redumped = reconstructed.model_dump()
            assert dumped == redumped, (
                "Serialization roundtrip failed! "
                "Check that `model_dump()` and `model_validate()` are inverses of each other."
            )
        except Exception as e:
            raise ValueError(f"Serialization roundtrip failed! {e}")

    def test_pickle_roundtrip(self, dumped: dict):
        try:
            pickled = pickle.dumps(dumped)
            unpickled = pickle.loads(pickled)
            assert dumped == unpickled, (
                "Pickle roundtrip failed! " "The unpickled object is not equal to the original."
            )
        except Exception as e:
            raise ValueError(f"Pickle serialization error: {e}")

    def test_serialization(self):
        dumped = self.model_dump()
        self.test_serialized_to_json(dumped)
        self.test_serialization_roundtrip(dumped)
        self.test_pickle_roundtrip(dumped)


SerializableType = BaseModel | Serializable | str | int | float | bool | dict | list


class TaskInput(BaseModel):
    args: tuple[SerializableType, ...]
    kwargs: dict[str, SerializableType]


class TaskOutput(BaseModel):
    return_value: SerializableType


def serialize_task_input(*args, **kwargs) -> dict:
    return TaskInput(args=args, kwargs=kwargs).model_dump()


def serialize_task_output(return_value) -> dict:
    return TaskOutput(return_value=return_value).model_dump()



# def score(self):
#     pass

# # todo - nest this in observation
# def write_text(self, file_name: str, text: str):
#     (self._runs_dir / file_name).write_text(text)

# def write_json(self, file_name: str, data: dict):
#     pass

# def log(self, message: str) -> None:
#     run = _run_context.get()
#     if run is None:  # not in evaluation mode
#         self._log.info(message)  # fallback to regular logging
#         return

#     stack = run.observation_stack
#     if not stack:
#         raise ValueError("Observation stack unexpectedly empty")
#     current_observation = stack[-1]
#     current_observation.add_log(message)

# def write(self, name: str, content: str) -> None:
#     run = _run_context.get()
#     if run is None:  # not in evaluation mode
#         self._log.warning("Attempting to write file outside of run context")
#         return

#     stack = run.observation_stack
#     if not stack:
#         raise ValueError("Observation stack unexpectedly empty")

#     current_observation = stack[-1]
#     current_observation.add_file(name, content)

#     filepath = run.dir / name
#     filepath.write_text(content)



from __future__ import annotations

from contextlib import contextmanager
from pathlib import Path
from typing import Optional

from joblib import Memory


class TaskCache:
    _instance: Optional[TaskCache] = None

    def __init__(self, cache_dir: Path):
        self.cache_dir = cache_dir
        self.memory = Memory(self.cache_dir, verbose=0)
        self.enabled = True

    @classmethod
    def get_instance(cls) -> Optional[TaskCache]:
        return cls._instance

    @classmethod
    @contextmanager
    def enable(cls, cache_dir: Path):
        previous_instance = cls._instance
        try:
            cls._instance = TaskCache(cache_dir)
            yield cls._instance
        finally:
            cls._instance = previous_instance

    @classmethod
    def cached(cls, func):
        def wrapper(*args, **kwargs):
            cache_instance = cls.get_instance()

            if cache_instance and cache_instance.enabled:
                cached_func = cache_instance.memory.cache(func)
                return cached_func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        return wrapper



import asyncio
import logging
import random
from typing import Any, Callable, ParamSpec, Sequence, Type, TypeVar

from pydantic import BaseModel
from tenacity import AsyncRetrying, RetryError, stop_after_attempt, wait_exponential

from agentlens.dataset import Dataset
from agentlens.provider import Message, Provider

logger = logging.getLogger(__name__)

T = TypeVar("T", bound=BaseModel)
D = TypeVar("D", bound=Dataset)
P = ParamSpec("P")
R = TypeVar("R")
RunT = TypeVar("RunT")


class AI:
    message: Type[Message] = Message
    _providers: dict[str, Provider]
    # todo - need mechanism for spawning generations under the run
    # lens: Lens

    def __init__(
        self,
        *,
        providers: Sequence[Provider] = (),
    ):
        self._providers = {provider.name: provider for provider in providers}

    def _create_messages(
        self,
        *,
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
    ) -> list[Message]:
        # check for invalid combinations
        if messages and (system or prompt):
            raise ValueError("Cannot specify both 'messages' and 'system'/'prompt'")

        # create messages if passed prompts
        if not messages:
            messages = []
            if system:
                messages.append(Message.system(system))
            if prompt:
                messages.append(Message.user(prompt))

        # apply dedent if needed
        return messages if not dedent else [m.dedent() for m in messages]

    def _parse_model_id(self, model_id: str) -> tuple[str, str]:
        """Transforms 'provider:model' -> (provider, model)"""
        try:
            provider_name, model_name = model_id.split(":", 1)
            return provider_name, model_name
        except ValueError:
            raise ValueError(
                f"Invalid model identifier '{model_id}'. " f"Expected format: 'provider:model'"
            )

    def _get_provider(self, model_id: str) -> tuple[Provider, str]:
        """Get the provider and parsed model name for a given model identifier"""
        provider_name, model_name = self._parse_model_id(model_id)

        provider = self._providers.get(provider_name)
        if not provider:
            raise ValueError(
                f"Provider '{provider_name}' not configured in AI manager. "
                f"Available providers: {list(self._providers.keys())}"
            )

        return provider, model_name

    async def _generate(
        self,
        generator: Callable,
        *,
        model: str,
        semaphore: asyncio.Semaphore,
        messages: list[Message] | None,
        system: str | None,
        prompt: str | None,
        dedent: bool,
        max_retries: int,
        **kwargs,
    ) -> Any:
        collected_messages = self._create_messages(
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
        )
        try:
            async for attempt in AsyncRetrying(
                stop=stop_after_attempt(max_retries),
                wait=wait_exponential(multiplier=1, min=1, max=10),
                reraise=True,
            ):
                with attempt:
                    try:
                        async with semaphore:
                            await asyncio.sleep(random.uniform(0, 0.1))
                            return await generator(
                                model=model,
                                messages=collected_messages,
                                **kwargs,
                            )
                    except Exception as e:
                        logger.debug(
                            f"Retry ({attempt.retry_state.attempt_number} of {max_retries}): {e}"
                        )
                        raise e
        except RetryError as e:
            logger.debug(f"Failed after {max_retries} attempts: {e}")
            raise e

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
        max_retries: int = 3,
        **kwargs,
    ) -> str:
        provider, model_name = self._get_provider(model)
        return await self._generate(
            provider.generate_text,
            model=model_name,
            semaphore=provider.get_semaphore(model_name),
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
            max_retries=max_retries,
            **kwargs,
        )

    async def generate_object(
        self,
        *,
        model: str,
        type: Type[T],
        messages: list[Message] | None = None,
        system: str | None = None,
        prompt: str | None = None,
        dedent: bool = True,
        max_retries: int = 3,
        **kwargs,
    ) -> T:
        # inline types may have invalid __name__ attributes -- replace w/ a default
        if hasattr(type, "__name__"):
            type.__name__ = "Response"
        provider, model_name = self._get_provider(model)
        return await self._generate(
            provider.generate_object,
            model=model_name,
            semaphore=provider.get_semaphore(model_name),
            type=type,
            messages=messages,
            system=system,
            prompt=prompt,
            dedent=dedent,
            max_retries=max_retries,
            **kwargs,
        )



from typing import Type, TypeVar

from anthropic import AsyncAnthropic
from pydantic import BaseModel

from agentlens.provider import Message, Provider

T = TypeVar("T", bound=BaseModel)


class AnthropicProvider(Provider):
    def __init__(
        self,
        api_key: str | None = None,
        max_connections: dict[str, int] = {"DEFAULT": 10},
    ):
        super().__init__(name="anthropic", max_connections=max_connections)
        self.client = AsyncAnthropic(api_key=api_key)

    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        **kwargs,
    ) -> str:
        completion = await self.client.messages.create(
            model=model,
            messages=[{"role": m.role, "content": m.content} for m in messages],
            **kwargs,
        )
        assert completion.content is not None
        return completion.content

    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        type: Type[T],
        **kwargs,
    ) -> T:
        raise NotImplementedError("Anthropic does not support object generation")



import asyncio
import logging
from abc import ABC, abstractmethod
from typing import Type, TypeVar

from agentlens.message import Message

logger = logging.getLogger(__name__)

T = TypeVar("T")


class Provider(ABC):
    def __init__(
        self,
        name: str,
        max_connections: dict[str, int] = {"DEFAULT": 10},
    ):
        self.name = name
        self._semaphores: dict[str, asyncio.Semaphore] = {}

        for model, limit in max_connections.items():
            self._semaphores[model] = asyncio.Semaphore(limit)

        self._default_semaphore = asyncio.Semaphore(max_connections.get("DEFAULT", 10))

    def get_semaphore(self, model: str) -> asyncio.Semaphore:
        return self._semaphores.get(model, self._default_semaphore)

    @abstractmethod
    async def generate_text(
        self,
        *,
        model: str,
        messages: list[Message],
        dedent: bool,
        max_retries: int,
        **kwargs,
    ) -> str:
        pass

    @abstractmethod
    async def generate_object(
        self,
        *,
        model: str,
        messages: list[Message],
        type: Type[T],
        **kwargs,
    ) -> T:
        pass



from .dataset import Dataset
from .hooks import GeneratorHook
from .inference import AI
from .lens import Lens
from .provider_openai import OpenAIProvider

__all__ = ["AI", "OpenAIProvider", "Dataset", "Lens", "GeneratorHook"]
